/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"
#include "d3d_quake.h"

// 256 might be more efficient for rendering, but it will mean that most lightmaps are hit in any given frame
// we might yet take it down to 64, but right now 128 seems a good balance...
#define LIGHTMAP_SIZE	128

cvar_t r_lerplightstyle ("r_lerplightstyle", "1", true);

// allow players to specify the frequency at which they want lightmaps updated
// value of 0 = always
cvar_t r_lightupdatefrequency ("r_lightupdatefrequency", "0", true);

float r_lightupdatetime = 0;
float r_lightupdateinterval = 0;

bool r_lightnoupdate = false;


void D3D_GetLightUpdateTimes (void)
{
	// get the interval to update at
	if (r_lightupdatefrequency.value <= 1)
		r_lightupdateinterval = 0;
	else
		r_lightupdateinterval = 1.0f / r_lightupdatefrequency.value;

	// accumulate time
	r_lightupdatetime += r_frametime;

	// flag if we need to update
	if (r_lightupdatetime >= r_lightupdateinterval)
	{
		r_lightupdatetime = 0;
		r_lightnoupdate = false;
	}
	else r_lightnoupdate = true;
}


typedef struct d3d_lightmap_s
{
	// the texture object
	LPDIRECT3DTEXTURE9 lm_Texture;

	// size of allocations
	byte allocated[LIGHTMAP_SIZE];

	// dirty region
	RECT DirtyRect;

	// locked texture rectangle
	D3DLOCKED_RECT LockedRect;

	// true if modified
	bool modified;

	// 1 or 2 (byte or short)
	int lmbytes;

	// next lightmap in the chain
	struct d3d_lightmap_s *next;
} d3d_lightmap_t;

d3d_lightmap_t *d3d_Lightmaps = NULL;

unsigned int d3d_blocklights[3][18 * 18];


void D3D_ReleaseLightmaps (void)
{
	for (d3d_lightmap_t *lm = d3d_Lightmaps; lm; lm = lm->next)
	{
		if (lm->lm_Texture)
		{
			// unlock, release and NULL
			lm->lm_Texture->UnlockRect (0);
			lm->lm_Texture->Release ();
			lm->lm_Texture = NULL;
		}

		// clear allocations
		memset (lm->allocated, 0, sizeof (lm->allocated));

		// flag as unmodified
		lm->modified = false;
	}
}


void D3D_CreateLightmapTexture (d3d_lightmap_t *lm)
{
	if (d3d_GlobalCaps.AllowA16B16G16R16)
	{
		// attempt to create it in 64 bit mode
		// there's no RGB version of D3DFMT_A16B16G16R16 so we must use BGR
		// (fixme - allow single component here where the source data is also single component)
		HRESULT hr = d3d_Device->CreateTexture
		(
			LIGHTMAP_SIZE,
			LIGHTMAP_SIZE,
			1,
			0,
			D3DFMT_A16B16G16R16,
			D3DPOOL_MANAGED,
			&lm->lm_Texture,
			NULL
		);

		if (SUCCEEDED (hr))
		{
			lm->lmbytes = 2;
			return;
		}
	}

	// attempt to create it in 32 bit mode
	// we also use BGR here to keep it consistent for simpler code
	HRESULT hr = d3d_Device->CreateTexture
	(
		LIGHTMAP_SIZE,
		LIGHTMAP_SIZE,
		1,
		0,
		D3DFMT_A8R8G8B8,
		D3DPOOL_MANAGED,
		&lm->lm_Texture,
		NULL
	);

	if (FAILED (hr))
	{
		Sys_Error ("D3D_CreateLightmapTexture: Failed to create a Lightmap Texture");
		return;
	}

	lm->lmbytes = 1;
}


d3d_lightmap_t *D3D_LMAllocBlock (int w, int h, byte *x, byte *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;
	d3d_lightmap_t *lm;

	for (lm = d3d_Lightmaps; lm; lm = lm->next)
	{
		best = LIGHTMAP_SIZE;

		for (i = 0; i < LIGHTMAP_SIZE - w; i++)
		{
			best2 = 0;

			for (j = 0; j < w; j++)
			{
				if (lm->allocated[i + j] >= best) break;
				if (lm->allocated[i + j] > best2) best2 = lm->allocated[i + j];
			}

			if (j == w)
			{
				// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > LIGHTMAP_SIZE)
			continue;

		for (i = 0; i < w; i++)
			lm->allocated[*x + i] = best + h;

		if (!lm->modified)
		{
			// create the lightmap texture
			D3D_CreateLightmapTexture (lm);

			// now we lock the texture rectangle with D3DLOCK_NO_DIRTY_UPDATE because we want to manage the dirty regions ourselves
			// we actually leave the entire texture locked during the level, only unlocking when we come to either reuse or
			// destroy the lightmap texture.
			lm->lm_Texture->LockRect (0, &lm->LockedRect, NULL, D3DLOCK_NO_DIRTY_UPDATE);

			// flag as being modified here
			lm->modified = true;
		}

		return lm;
	}

	// didn't find a lightmap so allocate a new one
	lm = (d3d_lightmap_t *) malloc (sizeof (d3d_lightmap_t));

	// link it in
	lm->next = d3d_Lightmaps;
	d3d_Lightmaps = lm;

	// no lightmap texture
	lm->lm_Texture = NULL;

	// texture isn't locked
	lm->modified = false;

	// clear allocations
	memset (lm->allocated, 0, sizeof (lm->allocated));

	// call recursively to get the new one we've just allocated
	return D3D_LMAllocBlock (w, h, x, y);
}


/*
===============
D3D_AddDynamicLights
===============
*/
void D3D_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	mtexinfo_t	*tex = surf->texinfo;

	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		// not hit by this light
		if (!(surf->dlightbits & (1 << lnum))) continue;

		rad = cl_dlights[lnum].radius;

		dist = DotProduct (cl_dlights[lnum].origin, surf->plane->normal) - surf->plane->dist;

		rad -= fabs(dist);
		minlight = cl_dlights[lnum].minlight;

		if (rad < minlight) continue;

		minlight = rad - minlight;

		for (i = 0; i < 3; i++)
			impact[i] = cl_dlights[lnum].origin[i] - surf->plane->normal[i] * dist;

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];
		
		for (t = 0; t < surf->tmax; t++)
		{
			td = local[1] - t * 16;

			if (td < 0) td = -td;

			for (s = 0; s < surf->smax; s++)
			{
				sd = local[0] - s * 16;

				if (sd < 0) sd = -sd;

				if (sd > td)
					dist = sd + (td >> 1);
				else
					dist = td + (sd >> 1);

				if (dist < minlight)
				{
					d3d_blocklights[0][t * surf->smax + s] += (rad - dist) * 256;
					d3d_blocklights[1][t * surf->smax + s] += (rad - dist) * 256;
					d3d_blocklights[2][t * surf->smax + s] += (rad - dist) * 256;
				}
			}
		}
	}
}


unsigned short D3D_AddBlockLight16 (unsigned int bl)
{
	// we can't just store it in as they can soometimes exceed 65535
	// this still gives 2^6 * GLQuake granularity, but it needs a modulate 4x blend
	return (bl > 131071) ? 65535 : bl >> 1;
}

byte D3D_AddBlockLight8 (unsigned int bl)
{
	// back to regular granularity
	// note - modulate 2x blend is needed here... i feel a global coming on...
	return (byte) (bl > 65280 ? 255 : bl >> 8);
}

void D3D_BuildLightMap (msurface_t *surf, int dataofs, int stride)
{
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	unsigned	*bl;

	d3d_lightmap_t *lm = (d3d_lightmap_t *) surf->d3d_Lightmap;

	surf->cached_dlight = (surf->dlightframe == r_framecount);

	size = surf->smax * surf->tmax;
	lightmap = surf->samples;

	// set to full bright if no light data
	if (r_fullbright.value || !cl.worldbrush->lightdata)
	{
		for (i = 0; i < size; i++)
		{
			d3d_blocklights[0][i] = 255 * 256;
			d3d_blocklights[1][i] = 255 * 256;
			d3d_blocklights[2][i] = 255 * 256;
		}

		goto store;
	}

	// clear to no light
	for (i = 0; i < size; i++)
	{
		d3d_blocklights[0][i] = 0;
		d3d_blocklights[1][i] = 0;
		d3d_blocklights[2][i] = 0;
	}

	// add all the lightmaps
	if (lightmap)
	{
		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
		{
			scale = d_lightstylevalue[surf->styles[maps]];
			surf->cached_light[maps] = scale;

			for (i = 0, j = 0; i < size; i++)
			{
				d3d_blocklights[0][i] += lightmap[j++] * scale;
				d3d_blocklights[1][i] += lightmap[j++] * scale;
				d3d_blocklights[2][i] += lightmap[j++] * scale;
			}

			// skip to next lightmap
			lightmap += (size * 3);
		}
	}

	// add all the dynamic lights
	if (surf->dlightframe == r_framecount)
		D3D_AddDynamicLights (surf);

store:;
	// retrieve pointers to the blocklights
	unsigned int *blr = d3d_blocklights[0];
	unsigned int *blg = d3d_blocklights[1];
	unsigned int *blb = d3d_blocklights[2];

	// both are 4 component
	stride -= (surf->smax << 2);

	if (lm->lmbytes == 1)
	{
		// 32 bit lightmap texture
		byte *dest = ((byte *) lm->LockedRect.pBits) + dataofs;

		for (i = 0; i < surf->tmax; i++, dest += stride)
		{
			for (j = 0; j < surf->smax; j++)
			{
				// locked in rgba format
				dest[0] = D3D_AddBlockLight8 (*blr++);
				dest[1] = D3D_AddBlockLight8 (*blg++);
				dest[2] = D3D_AddBlockLight8 (*blb++);
				dest[3] = 255;

				dest += 4;
			}
		}
	}
	else
	{
		// 64 bit lightmap texture
		unsigned short *dest = ((unsigned short *) lm->LockedRect.pBits) + dataofs;

		for (i = 0; i < surf->tmax; i++, dest += stride)
		{
			for (j = 0; j < surf->smax; j++)
			{
				// locked in rgba format
				dest[0] = D3D_AddBlockLight16 (*blr++);
				dest[1] = D3D_AddBlockLight16 (*blg++);
				dest[2] = D3D_AddBlockLight16 (*blb++);
				dest[3] = 65535;

				dest += 4;
			}
		}
	}
}


void D3D_CreateSurfaceLightmap (msurface_t *surf)
{
	if (surf->flags & SURF_DRAWSKY) return;
	if (surf->flags & SURF_DRAWTURB) return;

	// fill in lightmap extents
	surf->smax = (surf->extents[0] >> 4) + 1;
	surf->tmax = (surf->extents[1] >> 4) + 1;

	// create a lightmap for the surf (also fills in light_l and light_t
	surf->d3d_Lightmap = D3D_LMAllocBlock (surf->smax, surf->tmax, &surf->light_l, &surf->light_t);

	// fill in lightmap right and bottom (these exist just because I'm lazy and don't want to add a few numbers during updates)
	surf->light_r = surf->light_l + surf->smax;
	surf->light_b = surf->light_t + surf->tmax;

	// data offset in components, not in bits or bytes!
	int dataofs = (surf->light_t * LIGHTMAP_SIZE + surf->light_l) * 4;

	// build the lightmap
	D3D_BuildLightMap (surf, dataofs, LIGHTMAP_SIZE * 4);
}


void D3D_UploadLightmaps (void)
{
	// upload all lightmaps
	for (d3d_lightmap_t *lm = d3d_Lightmaps; lm; lm = lm->next)
	{
		// because we only delete textures rather than d3d_lightmap_t structs we need to check this
		if (lm->modified)
		{
			// mark the entire texture as dirty
			lm->lm_Texture->AddDirtyRect (NULL);

			// tell D3D we're going to need this managed resource shortly
			lm->lm_Texture->PreLoad ();
		}

		// dirty region is nothing at the start
		lm->DirtyRect.left = LIGHTMAP_SIZE;
		lm->DirtyRect.right = 0;
		lm->DirtyRect.top = LIGHTMAP_SIZE;
		lm->DirtyRect.bottom = 0;

		// not modified by default
		lm->modified = false;
	}
}


void D3D_BindLightmap (int stage, void *lm)
{
	D3D_BindTexture (stage, ((d3d_lightmap_t *) lm)->lm_Texture);
}


void D3D_UnlockLightmaps (void)
{
	if (r_lightnoupdate) return;

	// note - this function is misnamed now...
	for (d3d_lightmap_t *lm = d3d_Lightmaps; lm; lm = lm->next)
	{
		// see was it modified
		if (lm->modified)
		{
			// add a dirty region for that which was modified
			lm->lm_Texture->AddDirtyRect (&lm->DirtyRect);

			// not modified now
			lm->modified = false;

			// reset dirty region
			lm->DirtyRect.left = LIGHTMAP_SIZE;
			lm->DirtyRect.right = 0;
			lm->DirtyRect.top = LIGHTMAP_SIZE;
			lm->DirtyRect.bottom = 0;
		}
	}
}


void D3D_ModifySurfaceLightmap (msurface_t *surf)
{
	// retrieve the lightmap
	d3d_lightmap_t *lm = (d3d_lightmap_t *) surf->d3d_Lightmap;

	// flag as modified
	lm->modified = true;

	// get the region to make dirty for this update
	// notice how D3D uses top as 0, and right and bottom rather than width and height, making this so much cleaner (== les bug-prone)
	if (surf->light_l < lm->DirtyRect.left) lm->DirtyRect.left = surf->light_l;
	if (surf->light_t < lm->DirtyRect.top) lm->DirtyRect.top = surf->light_t;
	if (surf->light_r > lm->DirtyRect.right) lm->DirtyRect.right = surf->light_r;
	if (surf->light_b > lm->DirtyRect.bottom) lm->DirtyRect.bottom = surf->light_b;

	// data offset in components, not in bits or bytes!
	int dataofs = (surf->light_t * LIGHTMAP_SIZE + surf->light_l) * 4;

	// rebuild the lightmap
	D3D_BuildLightMap (surf, dataofs, LIGHTMAP_SIZE * 4);
}


void D3D_CheckLightmapModification (msurface_t *surf)
{
	// no lightmaps
	if (surf->flags & SURF_DRAWSKY) return;
	if (surf->flags & SURF_DRAWTURB) return;

	// no lightmap modifications
	if (!r_dynamic.value) return;
	if (r_lightnoupdate) return;

	// check for lightmap modification - cached lightstyle change
	for (int maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
	{
		if (d_lightstylevalue[surf->styles[maps]] != surf->cached_light[maps])
		{
			D3D_ModifySurfaceLightmap (surf);
			return;
		}
	}

	// dynamic this frame || dynamic previous frame
	if (surf->dlightframe == r_framecount || surf->cached_dlight) D3D_ModifySurfaceLightmap (surf);
}


void D3D_MakeLightmapTexCoords (msurface_t *surf, float *v, float *st)
{
	st[0] = DotProduct (v, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
	st[0] -= surf->texturemins[0];
	st[0] += surf->light_l * 16;
	st[0] += 8;
	st[0] /= LIGHTMAP_SIZE * 16;

	st[1] = DotProduct (v, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
	st[1] -= surf->texturemins[1];
	st[1] += surf->light_t * 16;
	st[1] += 8;
	st[1] /= LIGHTMAP_SIZE * 16;
}


/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (void)
{
	// made this cvar-controllable!
	if (r_lerplightstyle.value)
	{
		// interpolated light animations
		int			j, k;
		float		l;
		int			flight;
		int			clight;
		float		lerpfrac;
		float		backlerp;

		// light animations
		// 'm' is normal light, 'a' is no light, 'z' is double bright
		flight = (int) floor (cl.time * 10);
		clight = (int) ceil (cl.time * 10);
		lerpfrac = (cl.time * 10) - flight;
		backlerp = 1.0f - lerpfrac;

		for (j = 0; j < MAX_LIGHTSTYLES; j++)
		{
			if (!cl_lightstyle[j].length)
			{
				// was 256, changed to 264 for consistency
				d_lightstylevalue[j] = 264;
				continue;
			}
			else if (cl_lightstyle[j].length == 1)
			{
				// single length style so don't bother interpolating
				d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
				continue;
			}

			// interpolate animating light
			// frame just gone
			k = flight % cl_lightstyle[j].length;
			k = cl_lightstyle[j].map[k] - 'a';
			l = (float) (k * 22) * backlerp;

			// upcoming frame
			k = clight % cl_lightstyle[j].length;
			k = cl_lightstyle[j].map[k] - 'a';
			l += (float) (k * 22) * lerpfrac;

			d_lightstylevalue[j] = (int) l;
		}
	}
	else
	{
		// old light animation
		int			i,j,k;

		i = (int)(cl.time*10);
		for (j=0 ; j<MAX_LIGHTSTYLES ; j++)
		{
			if (!cl_lightstyle[j].length)
			{
				d_lightstylevalue[j] = 264;
				continue;
			}

			k = i % cl_lightstyle[j].length;
			k = cl_lightstyle[j].map[k] - 'a';
			k = k*22;
			d_lightstylevalue[j] = k;
		}
	}
}

/*
=============================================================================

DYNAMIC LIGHTS BLEND RENDERING

=============================================================================
*/

typedef struct lightblend_s
{
	float xyz[3];
	DWORD c;
} lightblend_t;

lightblend_t blendverts[18];

void AddLightBlend (float r, float g, float b, float a2)
{
	float a;
	float b3 = (float) v_blend[3] / 255.0f;

	b3 = a = b3 + a2 * (1 - b3);
	v_blend[3] = b3 * 255.0f;

	a2 = a2 / a;

	v_blend[0] = v_blend[0] * (1 - a2) + r * a2 * 255.0f;
	v_blend[1] = v_blend[1] * (1 - a2) + g * a2 * 255.0f;
	v_blend[2] = v_blend[2] * (1 - a2) + b * a2 * 255.0f;
}


void R_RenderDlight (dlight_t *light)
{
	int		i, j;
	float	a;
	vec3_t	v;
	float	rad;

	rad = light->radius * 0.35;

	VectorSubtract (light->origin, r_origin, v);

	if (Length (v) < rad)
	{
		// view is inside the dlight
		AddLightBlend (1, 0.5, 0, light->radius * 0.0003);
		return;
	}

	for (i = 0; i < 3; i++)
		blendverts[0].xyz[i] = light->origin[i] - vpn[i] * rad;

	blendverts[0].c = D3DCOLOR_XRGB (BYTE_CLAMP (50), BYTE_CLAMP (32), BYTE_CLAMP (0));

	int numverts = 1;

	for (i = 16; i >= 0; i--)
	{
		a = i / 16.0 * M_PI * 2;

		for (j = 0; j < 3; j++)
			blendverts[numverts].xyz[j] = light->origin[j] + vright[j] * cos (a) * rad + vup[j] * sin (a) * rad;

		blendverts[numverts].c = 0xff000000;
		numverts++;
	}

	d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLEFAN, 16, blendverts, sizeof (lightblend_t));
}


/*
=============
R_RenderDlights
=============
*/
void R_RenderDlights (void)
{
	int i;
	dlight_t *l;

	if (!gl_flashblend.value) return;

	// synamic lights blend state
	d3d_Device->SetRenderState (D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
	d3d_Device->SetTextureStageState (0, D3DTSS_COLOROP, D3DTOP_DISABLE);
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, FALSE);
	d3d_Device->SetRenderState (D3DRS_ALPHABLENDENABLE, TRUE);
	d3d_Device->SetRenderState (D3DRS_BLENDOP, D3DBLENDOP_ADD);
	d3d_Device->SetRenderState (D3DRS_SRCBLEND, D3DBLEND_ONE);
	d3d_Device->SetRenderState (D3DRS_DESTBLEND, D3DBLEND_ONE);

	D3D_SetFVFStateManaged (D3DFVF_XYZ | D3DFVF_DIFFUSE);

	l = cl_dlights;

	for (i = 0; i < MAX_DLIGHTS; i++, l++)
	{
		if (l->die < cl.time || !l->radius)
			continue;

		R_RenderDlight (l);
	}

	d3d_Device->SetRenderState (D3DRS_SHADEMODE, D3DSHADE_FLAT);
	d3d_Device->SetTextureStageState (0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, TRUE);
	d3d_DisableAlphaBlend->Apply ();
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights
=============
*/
void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i;
	int			sidebit;

	if (node->contents < 0) return;

	splitplane = node->plane;
	dist = DotProduct (light->origin, splitplane->normal) - splitplane->dist;

	if (dist > light->radius)
	{
		R_MarkLights (light, bit, node->children[0]);
		return;
	}

	if (dist < -light->radius)
	{
		R_MarkLights (light, bit, node->children[1]);
		return;
	}

	// mark the polygons
	surf = cl.worldbrush->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// no lights on these
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// skip backfacing surfs
		if ((dist = DotProduct (light->origin, surf->plane->normal) - surf->plane->dist) >= 0)
			sidebit = 0;
		else sidebit = SURF_PLANEBACK;

		if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

		if (surf->dlightframe != r_framecount)
		{
			surf->dlightbits = 0;
			surf->dlightframe = r_framecount;
		}

		surf->dlightbits |= bit;
	}

	R_MarkLights (light, bit, node->children[0]);
	R_MarkLights (light, bit, node->children[1]);
}


/*
=============
R_PushDlights
=============
*/
void R_PushDlights (mnode_t *headnode)
{
	int		i;
	dlight_t	*l;

	if (gl_flashblend.value) return;

	l = cl_dlights;

	for (i = 0; i < MAX_DLIGHTS; i++, l++)
	{
		if (l->die < cl.time || !l->radius)
			continue;

		R_MarkLights (l, 1 << i, headnode);
	}
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;

bool R_RecursiveLightPoint (int *color, mnode_t *node, vec3_t start, vec3_t end)
{
	int			r;
	float		front, back, frac;
	int			side;
	mplane_t	*plane;
	vec3_t		mid;
	msurface_t	*surf;
	int			s, t, ds, dt;
	int			i;
	mtexinfo_t	*tex;
	byte		*lightmap;
	unsigned	scale;
	int			maps;

	// hit nothing
	if (node->contents < 0) return false;

	// calculate mid point
	plane = node->plane;
	front = DotProduct (start, plane->normal) - plane->dist;
	back = DotProduct (end, plane->normal) - plane->dist;
	side = front < 0;

	if ((back < 0) == side) return R_RecursiveLightPoint (color, node->children[side], start, end);

	frac = front / (front - back);

	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side	
	if (R_RecursiveLightPoint (color, node->children[side], start, mid)) return true;

	// hit nothing
	if ((back < 0) == side) return false;

	// check for impact on this node
	VectorCopy (mid, lightspot);
	lightplane = plane;

	surf = cl.worldbrush->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// no lightmaps
		if (surf->flags & SURF_DRAWTILED) continue;

		tex = surf->texinfo;

		s = DotProduct (mid, tex->vecs[0]) + tex->vecs[0][3];
		t = DotProduct (mid, tex->vecs[1]) + tex->vecs[1][3];;

		if (s < surf->texturemins[0] || t < surf->texturemins[1]) continue;

		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];

		if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

		if (!surf->samples) return false;

		ds >>= 4;
		dt >>= 4;

		// note - already checked for (!surf->samples) above so this is guaranteed to have a value
		lightmap = surf->samples;

		// * 3 for coloured expansion
		lightmap += ((dt * surf->smax) + ds) * 3;

		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
		{
			color[0] += lightmap[0] * d_lightstylevalue[surf->styles[maps]];
			color[1] += lightmap[1] * d_lightstylevalue[surf->styles[maps]];
			color[2] += lightmap[2] * d_lightstylevalue[surf->styles[maps]];

			// * 3 for coloured expansion
			lightmap += surf->smax * surf->tmax * 3;
		}

		return true;
	}

	// go down back side
	return R_RecursiveLightPoint (color, node->children[!side], mid, end);
}


void R_LightPoint (entity_t *e, int *c)
{
	vec3_t		start;
	vec3_t		end;
	int			lnum;
	int			add;
	vec3_t		dist;

	if (!cl.worldbrush->lightdata)
	{
		// no light data
		c[0] = c[1] = c[2] = 255;
		return;
	}

	// set start point
	start[0] = e->origin[0];
	start[1] = e->origin[1];
	start[2] = e->origin[2];

	// set end point
	end[0] = e->origin[0];
	end[1] = e->origin[1];
	end[2] = e->origin[2] - 4096;

	// initially nothing
	c[0] = c[1] = c[2] = 0;

	// get lighting
	R_RecursiveLightPoint (c, cl.worldbrush->nodes, start, end);

	// pull back the scale because we're multiplying by d_lightstylevalue above
	c[0] >>= 8;
	c[1] >>= 8;
	c[2] >>= 8;

	// add dynamic lights
	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (e->origin, cl_dlights[lnum].origin, dist);

			add = (cl_dlights[lnum].radius - Length (dist));

			if (add > 0)
			{
				c[0] += add;
				c[1] += add;
				c[2] += add;
			}
		}
	}

	// pull back again on account of the modulate2x blend
	c[0] >>= 1;
	c[1] >>= 1;
	c[2] >>= 1;
}

