

#include "quakedef.h"
#include "d3d_quake.h"
#include "resource.h"

// used for generating md5 hashes
#include <wincrypt.h>

LPDIRECT3DTEXTURE9 d3d_CurrentTexture[8] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
d3d_texture_t *d3d_Textures = NULL;

void D3D_ReleaseLightmaps (void);

// other textures we use
extern LPDIRECT3DTEXTURE9 playertextures[];
extern LPDIRECT3DTEXTURE9 char_texture;
extern LPDIRECT3DTEXTURE9 solidskytexture;
extern LPDIRECT3DTEXTURE9 alphaskytexture;

// external textures directory - note - this should be registered before any textures are loaded, otherwise we will attempt to load
// them from /textures
cvar_t gl_texturedir ("gl_texturedir", "textures", true);

// mipmap tables
int ls_up_table[256];
byte ls_dn_table[200000];

void D3D_InitTextures (void)
{
	int i;

	// build the fast mipmap gamma tables
	for (i = 0; i < 256; i++)
	{
		float f;

		ls_up_table[i] = pow ((float) i, 2.2f);
	}

	for (i = 0; i < 200000; i++)
	{
		float f = pow ((float) i, 0.454545f) + 0.5f;;

		ls_dn_table[i] = (byte) (f > 255 ? 255 : f);
	}
}


void D3D_BindTexture (int stage, LPDIRECT3DTEXTURE9 tex)
{
	// no change
	if (tex == d3d_CurrentTexture[stage]) return;

	// set the texture
	d3d_Device->SetTexture (stage, tex);

	// store back to current
	d3d_CurrentTexture[stage] = tex;
}


void D3D_BindTexture (LPDIRECT3DTEXTURE9 tex)
{
	D3D_BindTexture (D3D_TEXTURE0, tex);
}


void D3D_HashTexture (image_t *image)
{
	// generate an MD5 hash of an image's data
	HCRYPTPROV hCryptProv;
	HCRYPTHASH hHash;

	// acquire the cryptographic context
	if (CryptAcquireContext (&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_MACHINE_KEYSET))
	{
		// create a hashing algorithm
		if (CryptCreateHash (hCryptProv, CALG_MD5, 0, 0, &hHash))
		{
			int datalen = image->height * image->width;

			if (image->flags & IMAGE_32BIT) datalen *= 4;

			// hash the data
			if (CryptHashData (hHash, image->data, datalen, 0))
			{
				DWORD dwHashLen = 16;

				// retrieve the hash
				if (CryptGetHashParam (hHash, HP_HASHVAL, image->hash, &dwHashLen, 0)) 
				{
					// hashed OK
				}
			}

			CryptDestroyHash (hHash); 
		}

	    CryptReleaseContext (hCryptProv, 0);
	}
}


void D3D_ResampleTexture (image_t *src, image_t *dst)
{
	// take an unsigned pointer to the dest data that we'll actually fill
	unsigned int *dstdata = (unsigned int *) dst->data;

	// easier access to src data for 32 bit resampling
	unsigned int *srcdata = (unsigned int *) src->data;

	// nearest neighbour for now
	for (int y = 0, dstpos = 0; y < dst->height; y++)
	{
		int srcbase = (y * src->height / dst->height) * src->width;

		for (int x = 0; x < dst->width; x++, dstpos++)
		{
			int srcpos = srcbase + (x * src->width / dst->width);

			if (src->flags & IMAGE_32BIT)
				dstdata[dstpos] = srcdata[srcpos];
			else if (src->palette)
				dstdata[dstpos] = src->palette[src->data[srcpos]];
			else Sys_Error ("D3D_ResampleTexture: !(flags & IMAGE_32BIT) without palette set");
		}
	}
}


void D3D_MipMap (byte *in, byte *out, int width, int height)
{
	int i, j;

	width <<= 2;
	height >>= 1;

	for (i = 0; i < height; i++, in += width)
	{
		for (j = 0; j < width; j += 8, out += 4, in += 8)
		{
			byte *in2 = &in[width];

			out[0] = ls_dn_table[((ls_up_table[in[0]] + ls_up_table[in[4]] + ls_up_table[in2[0]] + ls_up_table[in2[4]]) >> 2)];
			out[1] = ls_dn_table[((ls_up_table[in[1]] + ls_up_table[in[5]] + ls_up_table[in2[1]] + ls_up_table[in2[5]]) >> 2)];
			out[2] = ls_dn_table[((ls_up_table[in[2]] + ls_up_table[in[6]] + ls_up_table[in2[2]] + ls_up_table[in2[6]]) >> 2)];

			// leave alpha as is
			out[3] = (in[3] + in[7] + in2[3] + in2[7]) >> 2;
		}
	}
}


// types we're going to support - NOTE - link MUST be first in this list!
char *TextureExtensions[] = {"link", "dds", "tga", "bmp", "png", "jpg", NULL};

bool D3D_LoadExternalTexture (LPDIRECT3DTEXTURE9 *tex, char *filename, int flags)
{
	// dedicated servers go through here so don't crash them
	if (isDedicated) return false;

	char workingname[256];

	// copy the name out so that we can safely modify it
	strncpy (workingname, filename, 255);

	// ensure that we have some kind of extension on it - anything will do
	COM_DefaultExtension (workingname, ".blah");

	// change * to #
	for (int c = 0; ; c++)
	{
		if (workingname[c] == 0) break;
		if (workingname[c] == '*') workingname[c] = '#';
	}

	for (int i = 0; ; i++)
	{
		// ran out of extensions
		if (!TextureExtensions[i]) break;

		// set the correct extension
		for (int c = strlen (workingname) - 1; c; c--)
		{
			if (workingname[c] == '.')
			{
				strcpy (&workingname[c + 1], TextureExtensions[i]);
				break;
			}
		}

		FILE *f;
		int filelen = COM_FOpenFile (workingname, &f);

		if (!f) continue;

		if (i == 0)
		{
			// handle link format
			for (int c = strlen (workingname) - 1; c; c--)
			{
				if (workingname[c] == '/' || workingname[c] == '\\')
				{
					fscanf (f, "%s", &workingname[c + 1]);
					break;
				}
			}

			// done with the file
			fclose (f);

			// skip the rest; we check link first so that it will fall through to the other standard types
			continue;
		}

		// allocate a buffer to hold it
		// put this on fast hunk as it's already been allocated and will be filled in immediately
		int start = Hunk_LowMark ();
		byte *filebuf = (byte *) Hunk_AllocFast (filelen);

		// read it all in
		fread (filebuf, filelen, 1, f);

		// done with the file
		fclose (f);

		// attempt to load it
		HRESULT hr = D3DXCreateTextureFromFileInMemoryEx
		(
			d3d_Device,
			filebuf,
			filelen,
			D3DX_DEFAULT,
			D3DX_DEFAULT,
			(flags & IMAGE_MIPMAP) ? D3DX_DEFAULT : 1,
			0,
			D3DFMT_UNKNOWN,
			D3DPOOL_MANAGED,
			D3DX_FILTER_TRIANGLE | D3DX_FILTER_DITHER,
			D3DX_FILTER_BOX,
			0,
			NULL,
			NULL,
			tex
		);

		// done with the buffer
		Hunk_FreeToLowMark (start);

		// failed to load
		if (FAILED (hr)) continue;

		// load succeeded
		return true;
	}

	// didn't find it
	return false;
}


void D3D_LoadResourceTexture (LPDIRECT3DTEXTURE9 *tex, int ResourceID, int flags)
{
	// dedicated servers go through here so don't crash them
	if (isDedicated) return;

	HRESULT hr = D3DXCreateTextureFromResourceEx
	(
		d3d_Device,
		NULL,
		MAKEINTRESOURCE (ResourceID),
		D3DX_DEFAULT,
		D3DX_DEFAULT,
		(flags & IMAGE_MIPMAP) ? D3DX_DEFAULT : 1,
		0,
		D3DFMT_UNKNOWN,
		D3DPOOL_MANAGED,
		D3DX_FILTER_TRIANGLE | D3DX_FILTER_DITHER,
		D3DX_FILTER_BOX,
		0,
		NULL,
		NULL,
		tex
	);

	if (FAILED (hr))
	{
		// a resource texture failing is a program crash bug
		Sys_Error ("D3D_LoadResourceTexture: Failed to create texture");
		return;
	}

	// not much more we need to do here...
}


void D3D_LoadTexture (LPDIRECT3DTEXTURE9 *tex, image_t *image)
{
	// dedicated servers go through here so don't crash them
	if (isDedicated) return;

	// scaled image
	image_t scaled;

	// check scaling here first
	for (scaled.width = 1; scaled.width < image->width; scaled.width *= 2);
	for (scaled.height = 1; scaled.height < image->height; scaled.height *= 2);

	// clamp to max texture size
	if (scaled.width > d3d_DeviceCaps.MaxTextureWidth) scaled.width = d3d_DeviceCaps.MaxTextureWidth;
	if (scaled.height > d3d_DeviceCaps.MaxTextureHeight) scaled.height = d3d_DeviceCaps.MaxTextureHeight;

	if (image->flags & IMAGE_LIQUID)
	{
		// ensure liquid textures are square
		if (scaled.width > scaled.height) scaled.height = scaled.width;
		if (scaled.height > scaled.width) scaled.width = scaled.height;
	}

	// create the texture at the scaled size
	HRESULT hr = d3d_Device->CreateTexture
	(
		scaled.width,
		scaled.height,
		(image->flags & IMAGE_MIPMAP) ? 0 : 1,
		0,
		(image->flags & IMAGE_ALPHA) ? D3DFMT_A8R8G8B8 : D3DFMT_X8R8G8B8,
		D3DPOOL_MANAGED,
		tex,
		NULL
	);

	if (FAILED (hr))
	{
		Sys_Error ("D3D_LoadTexture: d3d_Device->CreateTexture failed");
		return;
	}

	// the rectangle to lock
	D3DLOCKED_RECT LockRect;

	// lock the texture rectangle
	(*tex)->LockRect (0, &LockRect, NULL, 0);

	// fill it in - how we do it depends on the scaling
	if (scaled.width == image->width && scaled.height == image->height)
	{
		// no scaling
		for (int i = 0; i < (scaled.width * scaled.height); i++)
		{
			unsigned int p;

			// retrieve the correct texel - this will either be direct or a palette lookup
			if (image->flags & IMAGE_32BIT)
				p = ((unsigned *) image->data)[i];
			else if (image->palette)
				p = image->palette[image->data[i]];
			else Sys_Error ("D3D_LoadTexture: !(flags & IMAGE_32BIT) without palette set");

			// store it back
			((unsigned *) LockRect.pBits)[i] = p;
		}
	}
	else
	{
		// save out lockbits in scaled data pointer
		scaled.data = (byte *) LockRect.pBits;

		// resample data into the texture
		D3D_ResampleTexture (image, &scaled);
	}

	int unlocklevel = 0;

	// now either generate mipmaps or just unlock it
	if (image->flags & IMAGE_MIPMAP)
	{
		// texture rectangles to use for mipmapping
		D3DLOCKED_RECT MipRect;
		D3DLOCKED_RECT *CurrRect = &MipRect;
		D3DLOCKED_RECT *PrevRect = &LockRect;

		// because level 0 has already been done
		int miplevel = 1;

		while (scaled.width > 1 || scaled.height > 1)
		{
			// lock current
			(*tex)->LockRect (miplevel, CurrRect, NULL, 0);

			// mipmap the image data from the previous level into the current level
			D3D_MipMap ((byte *) PrevRect->pBits, (byte *) CurrRect->pBits, scaled.width, scaled.height);

			// unlock previous
			(*tex)->UnlockRect (miplevel - 1);

			// swap rectangles
			D3DLOCKED_RECT *TempRect = CurrRect;
			CurrRect = PrevRect;
			PrevRect = TempRect;

			// take down to next level
			scaled.width >>= 1;
			scaled.height >>= 1;

			// never go < 1
			if (scaled.width < 1) scaled.width = 1;
			if (scaled.height < 1) scaled.height = 1;

			// go to next level
			miplevel++;
		}

		// this will leave us with a locked rect when we come out, so we must unlock it
		// the level to unlock is 1 less than miplevel, as miplevel was incremented after locking
		unlocklevel = miplevel - 1;
	}

	// unlock the correct level
	(*tex)->UnlockRect (unlocklevel);

	// tell Direct 3D that we're going to be needing to use this managed resource shortly
	(*tex)->PreLoad ();
}


void D3D_LoadTexture (LPDIRECT3DTEXTURE9 *tex, int width, int height, byte *data, unsigned int *palette, bool mipmap, bool alpha)
{
	// create an image struct for it
	image_t image;

	image.data = data;
	image.flags = 0;
	image.height = height;
	image.width = width;
	image.palette = palette;

	if (mipmap) image.flags |= IMAGE_MIPMAP;
	if (alpha) image.flags |= IMAGE_ALPHA;
	if (!palette) image.flags |= IMAGE_32BIT;

	// upload direct without going through the texture cache, as we already have a texture object for this
	D3D_LoadTexture (tex, &image);
}


// checksums of textures that need hacking...
byte ShotgunShells[] = {202, 6, 69, 163, 17, 112, 190, 234, 102, 56, 225, 242, 212, 175, 27, 187};
byte plat_top1_cable[] = {209, 191, 162, 164, 213, 63, 224, 73, 227, 251, 229, 137, 43, 60, 25, 138};
byte plat_top1_bolt[] = {52, 114, 210, 88, 38, 70, 116, 171, 89, 227, 115, 137, 102, 79, 193, 35};
byte metal5_2_arc[] = {35, 233, 88, 189, 135, 188, 152, 69, 221, 125, 104, 132, 51, 91, 22, 15};
byte metal5_2_x[] = {207, 93, 199, 54, 82, 58, 152, 177, 67, 18, 185, 231, 214, 4, 164, 99};
byte metal5_4_arc[] = {27, 95, 227, 196, 123, 235, 244, 145, 211, 222, 14, 190, 37, 255, 215, 107};
byte metal5_4_double[] = {47, 119, 108, 18, 244, 34, 166, 42, 207, 217, 179, 201, 114, 166, 199, 35};
byte metal5_8_back[] = {199, 119, 111, 184, 133, 111, 68, 52, 169, 1, 239, 142, 2, 233, 192, 15};
byte metal5_8_rune[] = {60, 183, 222, 11, 163, 158, 222, 195, 124, 161, 201, 158, 242, 30, 134, 28};
byte sky_blue_solid[] = {220, 67, 132, 212, 3, 131, 54, 160, 135, 4, 5, 86, 79, 146, 123, 89};
byte sky_blue_alpha[] = {163, 123, 35, 117, 154, 146, 68, 92, 141, 70, 253, 212, 187, 18, 112, 149};
byte sky_purp_solid[] = {196, 173, 196, 177, 19, 221, 134, 159, 208, 159, 158, 4, 108, 57, 10, 108};
byte sky_purp_alpha[] = {143, 106, 19, 206, 242, 171, 137, 86, 161, 74, 156, 217, 85, 10, 120, 149};


LPDIRECT3DTEXTURE9 D3D_LoadTexture (image_t *image)
{
	// take a hash of the image data
	D3D_HashTexture (image);

	// look for a match
	for (d3d_texture_t *t = d3d_Textures; t; t = t->next)
	{
		// it's not beyond the bounds of possibility that we might have 2 textures with the same
		// data but different usage, so here we check for it and accomodate it
		if (image->flags != t->TexImage.flags) continue;

		// compare the hash and reuse if it matches
		if (!memcmp (image->hash, t->TexImage.hash, 16))
		{
			// set last usage to 0
			t->LastUsage = 0;

			// return it
			return t->d3d_Texture;
		}
	}

	// create a new one
	d3d_texture_t *tex = (d3d_texture_t *) malloc (sizeof (d3d_texture_t));

	// link it in
	tex->next = d3d_Textures;
	d3d_Textures = tex;

	// fill in the struct
	tex->LastUsage = 0;
	tex->d3d_Texture = NULL;

	// copy the image
	memcpy (&tex->TexImage, image, sizeof (image_t));

	// check for textures that have the same name but different data, and amend the name by the QRP standard
	if (!memcmp (image->hash, plat_top1_cable, 16))
		strcat (image->identifier, "_cable");
	else if (!memcmp (image->hash, plat_top1_bolt, 16))
		strcat (image->identifier, "_bolt");
	else if (!memcmp (image->hash, metal5_2_arc, 16))
		strcat (image->identifier, "_arc");
	else if (!memcmp (image->hash, metal5_2_x, 16))
		strcat (image->identifier, "_x");
	else if (!memcmp (image->hash, metal5_4_arc, 16))
		strcat (image->identifier, "_arc");
	else if (!memcmp (image->hash, metal5_4_double, 16))
		strcat (image->identifier, "_double");
	else if (!memcmp (image->hash, metal5_8_back, 16))
		strcat (image->identifier, "_back");
	else if (!memcmp (image->hash, metal5_8_rune, 16))
		strcat (image->identifier, "_rune");
	else if (!memcmp (image->hash, sky_blue_solid, 16))
		strcpy (image->identifier, "sky4_solid");
	else if (!memcmp (image->hash, sky_blue_alpha, 16))
		strcpy (image->identifier, "sky4_alpha");
	else if (!memcmp (image->hash, sky_purp_solid, 16))
		strcpy (image->identifier, "sky1_solid");
	else if (!memcmp (image->hash, sky_purp_alpha, 16))
		strcpy (image->identifier, "sky1_alpha");

	// hack the colour for certain models
	// fix white line at base of shotgun shells box
	if (!memcmp (image->hash, ShotgunShells, 16))
		memcpy (image->data, image->data + 32 * 31, 32);

	COM_CheckContentDirectory (&gl_texturedir, false);

	if (!D3D_LoadExternalTexture (&tex->d3d_Texture, va ("%s/%s", gl_texturedir.string, image->identifier), image->flags))
	{
		// upload through direct 3d
		D3D_LoadTexture (&tex->d3d_Texture, image);
	}

	// return the texture we got
	return tex->d3d_Texture;
}


LPDIRECT3DTEXTURE9 D3D_LoadTexture (char *identifier, int width, int height, byte *data, bool mipmap, bool alpha)
{
	image_t image;

	image.data = data;
	image.flags = 0;
	image.height = height;
	image.width = width;
	image.palette = d_8to24table;

	strcpy (image.identifier, identifier);

	if (mipmap) image.flags |= IMAGE_MIPMAP;
	if (alpha) image.flags |= IMAGE_ALPHA;

	// detect water textures
	if (identifier[0] == '*') image.flags |= IMAGE_LIQUID;

	return D3D_LoadTexture (&image);
}


LPDIRECT3DTEXTURE9 D3D_LoadTexture (int width, int height, byte *data, int flags)
{
	image_t image;

	image.data = data;
	image.flags = flags;
	image.height = height;
	image.width = width;
	image.palette = NULL;
	image.identifier[0] = 0;

	return D3D_LoadTexture (&image);
}


void D3D_ReleaseTextures (bool fullrelease = D3D_RELEASE_UNUSED)
{
	extern LPDIRECT3DTEXTURE9 d3d_MapshotTexture;

	for (d3d_texture_t *tex = d3d_Textures; tex; tex = tex->next)
	{
		SAFE_RELEASE (tex->d3d_Texture);
	}

	// release player textures
	for (int i = 0; i < 16; i++)
		SAFE_RELEASE (playertextures[i]);

	// release lightmaps too
	D3D_ReleaseLightmaps ();

	// other textures
	SAFE_RELEASE (char_texture);
	SAFE_RELEASE (solidskytexture);
	SAFE_RELEASE (alphaskytexture);
	SAFE_RELEASE (d3d_MapshotTexture);
}


