/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"
#include "d3d_quake.h"

extern	model_t	*loadmodel;

int		skytexturenum;

LPDIRECT3DTEXTURE9 solidskytexture = NULL;
LPDIRECT3DTEXTURE9 alphaskytexture = NULL;
LPDIRECT3DVERTEXBUFFER9 d3d_SkySphereVerts = NULL;

// theoretically every texture in a map could be a warp texture (e.g. an ammo box)
// to do - also precache the surfaces
LPDIRECT3DTEXTURE9 d3d_WarpUpdateTextures[MAX_MAP_TEXTURES] = {NULL};
int d3d_NumWarpUpdateTextures = 0;

void D3D_ReleaseWarpTextures (void)
{
	int NumRelease = 0;

	// the warp update textures are created in the default pool so we must be sure to 
	// release them if we have a device reset, and also on video shutdown
	for (int i = 0; i < MAX_MAP_TEXTURES; i++)
	{
		if (d3d_WarpUpdateTextures[i]) NumRelease++;

		SAFE_RELEASE (d3d_WarpUpdateTextures[i]);
	}

	Con_DPrintf ("Released %i Warp Update Textures\n", NumRelease);

	// ensure
	d3d_NumWarpUpdateTextures = 0;
}


typedef struct glwatervert_s
{
	float xyz[3];
	DWORD c;
	float st[2];
	float xyz2[3];
} glwatervert_t;

typedef struct glwaterpoly_s
{
	int numverts;
	glwatervert_t *verts;
} glwaterpoly_t;

typedef struct glskyvert_s
{
	// sky verts are only used for clipping the sky sphere, so they don't need texcoords
	float xyz[3];
	float xyz2[3];
} glskyvert_t;

typedef struct glskypoly_s
{
	int numverts;
	glskyvert_t *verts;
} glskypoly_t;

typedef struct skysphere_s
{
	float xyz[3];
	float st[2];
} skysphere_t;

msurface_t	*warpface;


/*
==============================================================================================================================

			RENDER TO TEXTURE WATER WARP UPDATING

	Thanks to http://www.two-kings.de/tutorials/dxgraphics/dxgraphics16.html for valuable info on this

==============================================================================================================================
*/

void D3D_CreateWarpUpdateTexture (texture_t *warptex)
{
	D3DSURFACE_DESC Desc;

	// get the description for level 0 of the original texture
	((LPDIRECT3DTEXTURE9) warptex->d3d_Texture)->GetLevelDesc (0, &Desc);

	// get width and height - we ensured that the liquid was (1) a power of 2, and (2) a square
	// when loading it, so we don't need to concern ourselves with that here.
	// note - the warp update texture is tiled 2x2, but it needs to go on a 4x4 size to retain the LOD of the original
	int update_width = Desc.Width * 4;
	int update_height = Desc.Height * 4;

	// clamp to max texture size
	// note - a 512x512 warp texture can easily hit the max size for an Intel 910 card (2048x2048)
	// hence we take it down another notch here to prevent too much of that happening!  (16 MB per texture!)
	if (update_width * 2 > d3d_DeviceCaps.MaxTextureWidth) update_width = d3d_DeviceCaps.MaxTextureWidth / 2;
	if (update_height * 2 > d3d_DeviceCaps.MaxTextureHeight) update_height = d3d_DeviceCaps.MaxTextureHeight / 2;

	// now - just in case the above 2 are weird on any card - take down to the lowest of the 2 to ensure we still have a square
	if (update_width > update_height) update_width = update_height;
	if (update_height > update_width) update_height = update_width;

	// now create the texture
	// RTT textures need to be in the default pool,
	// DON'T create these with D3DUSAGE_AUTOGENMIPMAP as the mipmap will autogen each time they're rendered to!!!
	// note 2 - as these are 16x the storage of the original we create them as 16 bit for conservation.
	// faster updates too...
	d3d_Device->CreateTexture
	(
		update_width,
		update_height,
		0,
		D3DUSAGE_RENDERTARGET,
		D3DFMT_R5G6B5,
		D3DPOOL_DEFAULT,
		&d3d_WarpUpdateTextures[d3d_NumWarpUpdateTextures],
		NULL
	);

	// store back to the texture struct
	warptex->d3d_WarpUpdateTexture = d3d_NumWarpUpdateTextures;

	// set width and height of the texture struct
	// done here to avoid the overhead of a GetDesc when updating
	warptex->width = update_width;
	warptex->height = update_height;
}


void D3D_InitWarpTextures (void)
{
	// dedicated servers might go through here so don't crash them
	if (isDedicated) return;

	// we might not be connected yet
	if (cls.state != ca_connected) return;

	// release any warp update textures we got
	D3D_ReleaseWarpTextures ();

	// start anew...
	d3d_NumWarpUpdateTextures = 0;

	// initialize warp update textures on all warp surfaces on all models
	for (int j = 1; j < MAX_MODELS; j++)
	{
		model_t *m = cl.model_precache[j];

		if (!m) break;
		if (m->name[0] == '*') continue;
		if (m->type != mod_brush) continue;

		for (int i = 0; i < m->bh->numtextures; i++)
		{
			// retrieve the texture
			texture_t *tex = m->bh->textures[i];

			// e2m3 gets this
			if (!tex) continue;

			// only do liquid textures
			if (tex->name[0] != '*') continue;

			if (d3d_NumWarpUpdateTextures == MAX_MAP_TEXTURES)
			{
				// to be honest, the map will have crashed long before we get to here, but it does seem prudent anyway
				Sys_Error ("D3D_InitWarpTextures: d3d_NumWarpUpdateTextures == MAX_MAP_TEXTURES");
				return;
			}

			// create the new texture for updates
			D3D_CreateWarpUpdateTexture (tex);

			// go to the next texture
			d3d_NumWarpUpdateTextures++;
		}
	}

	Con_DPrintf ("Created %i warp update textures\n", d3d_NumWarpUpdateTextures);
}


typedef struct rttverts_s
{
	float xyz[3];
	float rhw;
	float st[2];
} rttverts_t;

// we always draw 10 verts
rttverts_t rttverts[10];

float *warpsin = NULL;
float warptime;

// recreate gl_warp_sin.h
void R_InitWarp (void)
{
	int i;

	warpsin = (float *) malloc (256 * sizeof (float));

	for (i = 0; i < 256; i++)
	{
		float f = (float) i / 40.743665f;

		// just take the sin so that we can multiply it by a variable warp factor
		warpsin[i] = sin (f);
	}
}

__inline float R_WarpCalc (float s, float t, float warpmult)
{
	// bleeding-edge optimization!
	// calculate the warp as if the texture was 128 x 128
	return (s + warpsin[(int) ((t * 2) + warptime) & 255] * warpmult) * 0.015625f;
}

void D3D_UpdateWarpTexture (texture_t *tex)
{
	LPDIRECT3DSURFACE9 d3d_RenderSurface;
	LPDIRECT3DTEXTURE9 d3d_WarpUpdateTexture;

	// retrieve the texture for easier access
	d3d_WarpUpdateTexture = d3d_WarpUpdateTextures[tex->d3d_WarpUpdateTexture];

	// rather than having to clear the Z-buffer each time, instead we just disable writes to it
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, FALSE);

	// don't use state managed as it's a different render target!
	d3d_Device->SetFVF (D3DFVF_XYZRHW | D3DFVF_TEX1);

	// set as source texture
	d3d_Device->SetTexture (0, (LPDIRECT3DTEXTURE9) tex->d3d_Texture);

	float x, y, x2, warptess;
	float s, t, s2;

	// these are adequate as static values; there's really nothing much to be gained from
	// using higher tesselation factors, and r_waterwarp goes weird at different values
	// if you decrease r_warptess be sure to allocate more verts above!  (work it out yourself)
	float r_waterwarp = 8.0f;
	float r_warptess = 32.0f;

	// note - generating the miplevels by hand rendering to each level in the chain is *SUBSTANTIALLY* faster
	// than using D3DUSAGE_AUTOGENMIPMAP at creation time and GenerateMipSubLevels when updating.  Go figure.
	// perhaps GenerateMipSubLevels pulls the texture back into system memory and loads of jiggerypokery???
	// we ensured that both of these textures were square when generating so we can work off width alone here.
	// here's another area which defies what youmight think logical - we use the top level mip of the source
	// for all rendering operations as it's faster - you'd think setting LOD would be good.  Go figure again.
	for (int miplevel = 0, mipsize = tex->width; mipsize; miplevel++, mipsize >>= 1)
	{
		// calculate tesselation factor
		warptess = (float) mipsize / (128.0f / r_warptess);

		// get the surface to render to
		d3d_WarpUpdateTexture->GetSurfaceLevel (miplevel, &d3d_RenderSurface);

		// set it as the device render target
		d3d_Device->SetRenderTarget (0, d3d_RenderSurface);

		// output warp update tristrips
		for (x = 0.0f, s = 0.0f; s < 128.0f; x = x2, s = s2)
		{
			x2 = x + warptess;
			s2 = s + r_warptess;

			int numverts = 0;

			for (y = 0.0f, t = 0.0f; t <= 128.0f; y += warptess, t += r_warptess)
			{
				rttverts[numverts].xyz[0] = x;
				rttverts[numverts].xyz[1] = y;
				rttverts[numverts].xyz[2] = 0;

				rttverts[numverts].rhw = 1;

				rttverts[numverts].st[0] = R_WarpCalc (s, t, r_waterwarp);
				rttverts[numverts].st[1] = R_WarpCalc (t, s, r_waterwarp);

				numverts++;

				rttverts[numverts].xyz[0] = x2;
				rttverts[numverts].xyz[1] = y;
				rttverts[numverts].xyz[2] = 0;

				rttverts[numverts].rhw = 1;

				rttverts[numverts].st[0] = R_WarpCalc (s2, t, r_waterwarp);
				rttverts[numverts].st[1] = R_WarpCalc (t, s2, r_waterwarp);

				numverts++;
			}

			// draw the update
			d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLESTRIP, numverts - 2, rttverts, sizeof (rttverts_t));
		}

		// release the surface
		d3d_RenderSurface->Release ();
	}
}


void D3D_UpdateWarpTextures (void)
{
	// dedicated servers SHOULDN'T go through here but let's not take any chances
	if (isDedicated) return;

	// initialize the warp sin table (one time only)
	if (!warpsin) R_InitWarp ();

	// set the warp time (moving this calculation out of a loop)
	warptime = cl.time * 40.743665f;

	LPDIRECT3DSURFACE9 d3d_BackBuffer = NULL;

	// check all textures
	for (int i = 0; i < cl.worldbrush->numtextures; i++)
	{
		// retrieve the texture
		texture_t *tex = cl.worldbrush->textures[i];

		// e2m3 gets this
		if (!tex) continue;

		// note - we can't rely on texturechain here as inline bmodels don't use it!
		// instead we store a visframe for the texture and check that
		if (tex->visframe != r_framecount) continue;

		// water only!
		if (tex->name[0] != '*') continue;

		// if we haven't stored out the backbuffer yet we do so now
		if (!d3d_BackBuffer) d3d_Device->GetRenderTarget (0, &d3d_BackBuffer);

		// update the texture
		D3D_UpdateWarpTexture (tex);
	}

	// reset the backbuffer as the rendertarget
	if (d3d_BackBuffer)
	{
		// reset drawing params
		d3d_Device->SetRenderTarget (0, d3d_BackBuffer);
		d3d_BackBuffer->Release ();
	}
}


/*
==============================================================================================================================

		WARP SURFACE GENERATION

		No longer subdivided :D

==============================================================================================================================
*/

void GL_DontSubdivideSky (msurface_t *surf)
{
	int			i;
	int			lindex;
	float		*vec;

	if (surf->flags & SURF_DRAWTURB)
	{
		Sys_Error ("GL_DontSubdivideSky with (surf->flags & SURF_DRAWTURB)");
		return;
	}

	// alloc the poly in memory
	glskypoly_t *poly = (glskypoly_t *) Hunk_Alloc (sizeof (glskypoly_t));
	poly->verts = (glskyvert_t *) Hunk_Alloc (surf->numedges * sizeof (glskyvert_t));
	poly->numverts = surf->numedges;

	// link it into the chain (there is no chain!)
	surf->polys = poly;

	for (i = 0; i < poly->numverts; i++)
	{
		lindex = loadmodel->bh->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->bh->vertexes[loadmodel->bh->edges[lindex].v[0]].position;
		else
			vec = loadmodel->bh->vertexes[loadmodel->bh->edges[-lindex].v[1]].position;

		// take a second copy for transformations
		poly->verts[i].xyz[0] = poly->verts[i].xyz2[0] = vec[0];
		poly->verts[i].xyz[1] = poly->verts[i].xyz2[1] = vec[1];
		poly->verts[i].xyz[2] = poly->verts[i].xyz2[2] = vec[2];
	}
}


void GL_DontSubdivideWater (msurface_t *surf)
{
	int			i;
	int			lindex;
	float		*vec;

	if (surf->flags & SURF_DRAWSKY)
	{
		Sys_Error ("GL_DontSubdivideWater with (surf->flags & SURF_DRAWSKY)");
		return;
	}

	// alloc the poly in memory
	glwaterpoly_t *poly = (glwaterpoly_t *) Hunk_Alloc (sizeof (glwaterpoly_t));
	poly->verts = (glwatervert_t *) Hunk_Alloc (surf->numedges * sizeof (glwatervert_t));
	poly->numverts = surf->numedges;

	// link it into the chain (there is no chain!)
	surf->polys = poly;

	for (i = 0; i < poly->numverts; i++)
	{
		lindex = loadmodel->bh->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->bh->vertexes[loadmodel->bh->edges[lindex].v[0]].position;
		else
			vec = loadmodel->bh->vertexes[loadmodel->bh->edges[-lindex].v[1]].position;

		// take a second copy for transformations
		poly->verts[i].xyz[0] = poly->verts[i].xyz2[0] = vec[0];
		poly->verts[i].xyz[1] = poly->verts[i].xyz2[1] = vec[1];
		poly->verts[i].xyz[2] = poly->verts[i].xyz2[2] = vec[2];

		// texcoords (pre-scaled)
		poly->verts[i].st[0] = DotProduct (vec, surf->texinfo->vecs[0]) / 128.0f;
		poly->verts[i].st[1] = DotProduct (vec, surf->texinfo->vecs[1]) / 128.0f;
	}
}


/*
==============================================================================================================================

			SOFTWARE TRANSFORMATION FOR WARP SURFACES ON INLINE BMODELS

		The two different methods could be merged as the struct layouts are identical for the
		required data, although v++ might throw it a bit.  No big deal.

==============================================================================================================================
*/

void D3D_TransformWarpSurface (msurface_t *surf, D3DMATRIX *m)
{
	if (surf->flags & SURF_DRAWTURB)
	{
		glwaterpoly_t *p = (glwaterpoly_t *) surf->polys;
		glwatervert_t *v = p->verts;

		for (int i = 0; i < p->numverts; i++, v++)
		{
			// transform by the matrix then copy back
			v->xyz[0] = v->xyz2[0] * m->_11 + v->xyz2[1] * m->_21 + v->xyz2[2] * m->_31 + m->_41;
			v->xyz[1] = v->xyz2[0] * m->_12 + v->xyz2[1] * m->_22 + v->xyz2[2] * m->_32 + m->_42;
			v->xyz[2] = v->xyz2[0] * m->_13 + v->xyz2[1] * m->_23 + v->xyz2[2] * m->_33 + m->_43;
		}
	}
	else
	{
		glskypoly_t *p = (glskypoly_t *) surf->polys;
		glskyvert_t *v = p->verts;

		for (int i = 0; i < p->numverts; i++, v++)
		{
			// transform by the matrix then copy back
			v->xyz[0] = v->xyz2[0] * m->_11 + v->xyz2[1] * m->_21 + v->xyz2[2] * m->_31 + m->_41;
			v->xyz[1] = v->xyz2[0] * m->_12 + v->xyz2[1] * m->_22 + v->xyz2[2] * m->_32 + m->_42;
			v->xyz[2] = v->xyz2[0] * m->_13 + v->xyz2[1] * m->_23 + v->xyz2[2] * m->_33 + m->_43;
		}
	}
}


/*
==============================================================================================================================

		WATER WARP RENDERING

	Surface subdivision and warpsin updates are removed, water uses a dynamically updating warp texture.

==============================================================================================================================
*/


/*
=============
EmitWaterPolys

=============
*/
void EmitWaterPolys (msurface_t *surf, texture_t *tex, DWORD WaterDiffuse)
{
	glwaterpoly_t *p = (glwaterpoly_t *) surf->polys;
	glwatervert_t *v = p->verts;

	// bind the update texture
	D3D_BindTexture (0, d3d_WarpUpdateTextures[tex->d3d_WarpUpdateTexture]);

	// pass through the verts to update colour
	// there MUST be a way of doing this globally with a single call, but damned if i can figure it.
	// something in SetTextureStageState might be the ticket.  this will do for now
	for (int i = 0; i < p->numverts; i++, v++)
		v->c = WaterDiffuse;

	// draw the poly
	d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLEFAN, p->numverts - 2, p->verts, sizeof (glwatervert_t));
}


/*
==============================================================================================================================

		SKY WARP RENDERING

	Sky is rendered as a full sphere.  An initial "clipping" pass (writing to the depth buffer only) is made to
	establish the baseline area covered by the sky.  The old sky surfs are used for this
	
	The depth test is then inverted and the sky sphere is rendered as a "Z-fail" pass, meaning that areas where
	the depth buffer was rendered to in the previous pass are the only ones updated.

	A final clipping pass is made with the normal depth buffer to remove any remaining world geometry.
	(Note - this is currently commented out as I'm not certain it's actually necessary.  I remember it being needed
	when I wrote the code first, but that was a good while back and it was an earlier version of code.)

	This gives us a skysphere that is drawn at a distance but still clips world geometry.

	Note: you meed either an infinite projection matrix or a very large Z-far for this to work right!

==============================================================================================================================
*/

void R_ClipSky (msurface_t *skychain)
{
	// disable textureing and writes to the color buffer
	D3D_SetFVFStateManaged (D3DFVF_XYZ);
	d3d_Device->SetTextureStageState (0, D3DTSS_COLOROP, D3DTOP_DISABLE);
	d3d_Device->SetRenderState (D3DRS_COLORWRITEENABLE, 0);

	for (msurface_t *surf = skychain; surf; surf = surf->texturechain)
	{
		c_brush_polys++;

		// no longer subdivided
		glskypoly_t *p = (glskypoly_t *) surf->polys;

		// fixme - only need to store 3 floats for these - cut down on geometry going to the card
		d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLEFAN, p->numverts - 2, p->verts, sizeof (glskyvert_t));
	}

	// revert state
	d3d_Device->SetRenderState (D3DRS_COLORWRITEENABLE, 0x0000000F);
	d3d_Device->SetTextureStageState (0, D3DTSS_COLOROP, D3DTOP_MODULATE);
}


void R_DrawSkySphere (float rotatefactor, float scale, LPDIRECT3DTEXTURE9 skytexture)
{
	d3d_WorldMatrixStack->Push ();
	d3d_WorldMatrixStack->ScaleLocal (scale, scale, scale / 2);
	d3d_WorldMatrixStack->RotateAxisLocal (&XVECTOR, D3DXToRadian (90));
	d3d_WorldMatrixStack->RotateAxisLocal (&YVECTOR, D3DXToRadian (22));
	d3d_WorldMatrixStack->RotateAxisLocal (&ZVECTOR, D3DXToRadian (rotatefactor));
	d3d_Device->SetTransform (D3DTS_WORLD, d3d_WorldMatrixStack->GetTop ());

	D3D_BindTexture (skytexture);

	for (int i = 0; i < 220; i += 22)
		d3d_Device->DrawPrimitive (D3DPT_TRIANGLESTRIP, i, 20);

	d3d_WorldMatrixStack->Pop ();
}


/*
=================
R_DrawSkyChain
=================
*/
void R_DrawSkyChain (msurface_t *skychain)
{
	// no sky to draw!
	if (!skychain) return;

	// approximately replicate the speed of the old sky warp
	float rotateBack = anglemod (cl.time * 2.0f);
	float rotateFore = anglemod (cl.time * 4.0f);

	// write the regular sky polys into the depth buffer to get a baseline
	R_ClipSky (skychain);

	// flip the depth func so that the regular polys will prevent sphere polys outside their area reaching the framebuffer
	// also disable writing to Z
	d3d_Device->SetRenderState (D3DRS_ZFUNC, D3DCMP_GREATEREQUAL);
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, FALSE);

	// sky layer drawing state
	d3d_Device->SetStreamSource (0, d3d_SkySphereVerts, 0, sizeof (skysphere_t));
	D3D_SetFVFStateManaged (D3DFVF_XYZ | D3DFVF_TEX1);
	d3d_WorldMatrixStack->Push ();
	d3d_WorldMatrixStack->TranslateLocal (r_origin[0], r_origin[1], r_origin[2]);
	d3d_WorldMatrixStack->TranslateLocal (0, 0, -8000);

	// draw back layer
	R_DrawSkySphere (rotateBack, 10, solidskytexture);

	// draw front layer
	d3d_EnableAlphaBlend->Apply ();
	R_DrawSkySphere (rotateFore, 8, alphaskytexture);
	d3d_DisableAlphaBlend->Apply ();

	// restore the depth func
	d3d_Device->SetRenderState (D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, TRUE);

	// go back to the world matrix
	d3d_WorldMatrixStack->Pop ();
	d3d_WorldMatrixStack->LoadMatrix (&d3d_WorldMatrix);
	d3d_Device->SetTransform (D3DTS_WORLD, d3d_WorldMatrixStack->GetTop ());

	// now write the regular polys one more time to clip world geometry
	// not certain if this is actually necessary...!
	// i'll put it back it i notice any indication it's needed
	// R_ClipSky (skychain);
}


/*
==============================================================================================================================

		SKY INITIALIZATION

	The sphere is drawn from a VertexBuffer.  The sky texture is in two layers, top and bottom, contained in the same
	texture.

==============================================================================================================================
*/

void D3D_InitSkySphere (void)
{
	// already initialized
	if (d3d_SkySphereVerts) return;

	// set up the sphere vertex buffer
	HRESULT hr = d3d_Device->CreateVertexBuffer
	(
		220 * sizeof (skysphere_t),
		0,
		(D3DFVF_XYZ | D3DFVF_TEX1),
		D3DPOOL_MANAGED,
		&d3d_SkySphereVerts,
		NULL
	);

	// fixme - fall back on DrawPrimitiveUP if this happens!
	if (FAILED (hr)) Sys_Error ("D3D_InitSkySphere: d3d_Device->CreateVertexBuffer failed");

	skysphere_t *verts;

	// lock the full buffers
	d3d_SkySphereVerts->Lock (0, 0, (void **) &verts, 0);

	// fill it in
	float drho = 0.3141592653589;
	float dtheta = 0.6283185307180;

	float ds = 0.1;
	float dt = 0.1;

	float t = 1.0f;	
	float s = 0.0f;

	int i;
	int j;

	int vertnum = 0;

	for (i = 0; i < 10; i++)
	{
		float rho = (float) i * drho;
		float srho = (float) (sin (rho));
		float crho = (float) (cos (rho));
		float srhodrho = (float) (sin (rho + drho));
		float crhodrho = (float) (cos (rho + drho));

		s = 0.0f;

		for (j = 0; j <= 10; j++)
		{
			float theta = (j == 10) ? 0.0f : j * dtheta;
			float stheta = (float) (-sin (theta));
			float ctheta = (float) (cos (theta));

			verts[vertnum].st[0] = s * 24;
			verts[vertnum].st[1] = t * 12;

			verts[vertnum].xyz[0] = stheta * srho * 4096.0;
			verts[vertnum].xyz[1] = ctheta * srho * 4096.0;
			verts[vertnum].xyz[2] = crho * 4096.0;

			vertnum++;

			verts[vertnum].st[0] = s * 24;
			verts[vertnum].st[1] = (t - dt) * 12;

			verts[vertnum].xyz[0] = stheta * srhodrho * 4096.0;
			verts[vertnum].xyz[1] = ctheta * srhodrho * 4096.0;
			verts[vertnum].xyz[2] = crhodrho * 4096.0;

			vertnum++;
			s += ds;
		}

		t -= dt;
	}

	// unlock
	d3d_SkySphereVerts->Unlock ();
	d3d_SkySphereVerts->PreLoad ();
}


/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void R_InitSky (miptex_t *mt)
{
	// dedicated servers go through here so don't crash them
	if (isDedicated) return;

	// create the sky sphere (one time only)
	D3D_InitSkySphere ();

	int			i, j, p;
	byte		*src;
	unsigned	trans[128 * 128];
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;

	// destroy any current textures we might have
	SAFE_RELEASE (solidskytexture);
	SAFE_RELEASE (alphaskytexture);

	src = (byte *) mt + mt->offsets[0];

	// make an average value for the back to avoid
	// a fringe on the top level
	for (i = 0, r = 0, g = 0, b = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j + 128];

			rgba = &d_8to24table[p];

			trans[(i * 128) + j] = *rgba;

			r += ((byte *) rgba)[0];
			g += ((byte *) rgba)[1];
			b += ((byte *) rgba)[2];
		}
	}

	((byte *) &transpix)[0] = r / (128 * 128);
	((byte *) &transpix)[1] = g / (128 * 128);
	((byte *) &transpix)[2] = b / (128 * 128);
	((byte *) &transpix)[3] = 0;

	// upload it
	D3D_LoadTexture (&solidskytexture, 128, 128, (byte *) trans, NULL, false, false);

	// bottom layer
	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j];

			if (p == 0)
				trans[(i * 128) + j] = transpix;
			else
				trans[(i * 128) + j] = d_8to24table[p];
		}
	}

	// upload it
	D3D_LoadTexture (&alphaskytexture, 128, 128, (byte *) trans, NULL, false, true);
}

